@@ -25,6 +25,9 @@
 
 #include "common.h"
 #include "backendgvfs.h"
+#include "bookmark.h"
+#include "settings.h"
+#include "window.h"
 #include "mountoperation.h"
 
 typedef struct _GigoloBackendGVFSPrivate			GigoloBackendGVFSPrivate;
@@ -35,6 +38,7 @@ typedef struct _GigoloBackendGVFSPrivate			GigoloBackendGVFSPrivate;
 enum
 {
     PROP_0,
+    PROP_PARENT,
     PROP_STORE
 };
 
@@ -42,11 +46,21 @@ enum
 {
 	MOUNTS_CHANGED,
 	OPERATION_FAILED,
+	BROWSE_NETWORK_FINISHED,
+	BROWSE_HOST_FINISHED,
 
 	LAST_SIGNAL
 };
 static guint signals[LAST_SIGNAL];
 
+enum
+{
+    BROWSE_MODE_INVALID,
+    BROWSE_MODE_DOMAINS,
+    BROWSE_MODE_HOSTS,
+    BROWSE_MODE_SHARES,
+};
+
 typedef struct
 {
 	GigoloBackendGVFS *self;
@@ -54,14 +68,33 @@ typedef struct
 	gboolean show_errors;
 } MountInfo;
 
+typedef struct BrowseData
+{
+	GigoloBackendGVFS *self;
+
+	gint mode;
+	gchar *uri;
+	GtkWindow *parent;
+	GtkTreeStore *store;
+	GtkTreePath *parent_path;
+
+	void (*browse_func) (struct BrowseData *bd);
+} BrowseData;
+
+
 struct _GigoloBackendGVFSPrivate
 {
+	GtkWindow *parent;
 	GtkListStore *store;
+
+	gint browse_counter;
 };
 
 static void gigolo_backend_gvfs_finalize  			(GObject *object);
 static void gigolo_backend_gvfs_set_property		(GObject *object, guint prop_id,
 													 const GValue *value, GParamSpec *pspec);
+static void browse_network_real						(BrowseData *bd);
+static void browse_host_real						(BrowseData *bd);
 
 
 G_DEFINE_TYPE(GigoloBackendGVFS, gigolo_backend_gvfs, G_TYPE_OBJECT);
@@ -116,6 +149,15 @@ static void gigolo_backend_gvfs_class_init(GigoloBackendGVFSClass *klass)
 
 	g_type_class_add_private(klass, sizeof(GigoloBackendGVFSPrivate));
 
+	g_object_class_install_property(g_object_class,
+										PROP_PARENT,
+										g_param_spec_object(
+										"parent",
+										"Parent",
+										"Parent window",
+										GTK_TYPE_WINDOW,
+										G_PARAM_WRITABLE));
+
 	g_object_class_install_property(g_object_class,
 										PROP_STORE,
 										g_param_spec_object(
@@ -141,6 +183,23 @@ static void gigolo_backend_gvfs_class_init(GigoloBackendGVFSClass *klass)
 										NULL,
 										gigolo_backend_gvfs_cclosure_marshal_VOID__STRING_STRING,
 										G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);
+	signals[BROWSE_NETWORK_FINISHED] = g_signal_new("browse-network-finished",
+										G_TYPE_FROM_CLASS(klass),
+										(GSignalFlags) 0,
+										0,
+										0,
+										NULL,
+										g_cclosure_marshal_VOID__VOID,
+										G_TYPE_NONE, 0);
+	signals[BROWSE_HOST_FINISHED] = g_signal_new("browse-host-finished",
+										G_TYPE_FROM_CLASS(klass),
+										(GSignalFlags) 0,
+										0,
+										0,
+										NULL,
+										/* FIXME use a proper closure, arg1 is a GList* */
+										g_cclosure_marshal_VOID__POINTER,
+										G_TYPE_NONE, 1, G_TYPE_POINTER);
 }
 
 
@@ -154,17 +213,30 @@ static void gigolo_backend_gvfs_finalize(GObject *object)
 }
 
 
-static gchar *get_tooltip_text(gpointer ref, gint ref_type, const gchar *type)
+static gchar *get_tooltip_text(GigoloBackendGVFS *backend, gpointer ref, gint ref_type, const gchar *type)
 {
 	gchar *result = NULL;
+	GigoloBackendGVFSPrivate *priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(backend);
 	switch (ref_type)
 	{
 		case GIGOLO_WINDOW_REF_TYPE_MOUNT:
 		{
 			gchar *uri, *name, *clean_uri;
+			GigoloBookmark *b;
+			GigoloSettings *settings;
 
 			gigolo_backend_gvfs_get_name_and_uri_from_mount(ref, &name, &uri);
 			clean_uri = g_uri_unescape_string(uri, G_URI_RESERVED_CHARS_ALLOWED_IN_USERINFO);
+
+			settings = gigolo_window_get_settings(GIGOLO_WINDOW(priv->parent));
+			b = gigolo_settings_get_bookmark_by_uri(settings, uri);
+			if (b != NULL)
+			{
+				const gchar *folder = gigolo_bookmark_get_folder(b);
+				if (NZV(folder))
+					setptr(clean_uri, g_build_filename(clean_uri, folder, NULL));
+			}
+
 			result = g_strdup_printf(
 				_("<b>%s</b>\n\nURI: %s\nConnected: Yes\nService Type: %s"), name, clean_uri, type);
 
@@ -221,10 +293,9 @@ static void mount_volume_changed_cb(GVolumeMonitor *vm, G_GNUC_UNUSED GMount *mn
 		scheme_name = gigolo_describe_scheme(scheme);
 		uri = g_file_get_uri(file);
 		icon = g_mount_get_icon(mount);
-		tooltip_text = get_tooltip_text(mount, GIGOLO_WINDOW_REF_TYPE_MOUNT, scheme_name);
+		tooltip_text = get_tooltip_text(backend, mount, GIGOLO_WINDOW_REF_TYPE_MOUNT, scheme_name);
 
-		gtk_list_store_append(priv->store, &iter);
-		gtk_list_store_set(priv->store, &iter,
+		gtk_list_store_insert_with_values(priv->store, &iter, -1,
 				GIGOLO_WINDOW_COL_IS_MOUNTED, TRUE,
 				GIGOLO_WINDOW_COL_NAME, vol_name,
 				GIGOLO_WINDOW_COL_SCHEME, scheme_name,
@@ -255,10 +326,9 @@ static void mount_volume_changed_cb(GVolumeMonitor *vm, G_GNUC_UNUSED GMount *mn
 		{
 			icon = g_volume_get_icon(volume);
 			vol_name = g_volume_get_name(volume);
-			tooltip_text = get_tooltip_text(volume, GIGOLO_WINDOW_REF_TYPE_VOLUME, NULL);
+			tooltip_text = get_tooltip_text(backend, volume, GIGOLO_WINDOW_REF_TYPE_VOLUME, NULL);
 
-			gtk_list_store_append(priv->store, &iter);
-			gtk_list_store_set(priv->store, &iter,
+			gtk_list_store_insert_with_values(priv->store, &iter, -1,
 					GIGOLO_WINDOW_COL_IS_MOUNTED, FALSE,
 					GIGOLO_WINDOW_COL_NAME, vol_name,
 					GIGOLO_WINDOW_COL_SCHEME, gigolo_describe_scheme("file"),
@@ -283,14 +353,20 @@ static void mount_volume_changed_cb(GVolumeMonitor *vm, G_GNUC_UNUSED GMount *mn
 
 
 static void gigolo_backend_gvfs_set_property(GObject *object, guint prop_id,
-										   const GValue *value, GParamSpec *pspec)
+											 const GValue *value, GParamSpec *pspec)
 {
+	GigoloBackendGVFSPrivate *priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(object);
+
 	switch (prop_id)
 	{
+	case PROP_PARENT:
+	{
+		priv->parent = g_value_get_object(value);
+		break;
+	}
 	case PROP_STORE:
 	{
 		GVolumeMonitor *gvm;
-		GigoloBackendGVFSPrivate *priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(object);
 
 		priv->store = g_value_get_object(value);
 
@@ -318,11 +394,9 @@ static void gigolo_backend_gvfs_init(G_GNUC_UNUSED GigoloBackendGVFS *self)
 }
 
 
-GigoloBackendGVFS *gigolo_backend_gvfs_new(GtkListStore *store)
+GigoloBackendGVFS *gigolo_backend_gvfs_new(void)
 {
-	GigoloBackendGVFS *backend = g_object_new(GIGOLO_BACKEND_GVFS_TYPE, "store", store, NULL);
-
-	return backend;
+	return g_object_new(GIGOLO_BACKEND_GVFS_TYPE, NULL);
 }
 
 
@@ -368,13 +442,14 @@ static void volume_mount_finished_cb(GObject *src, GAsyncResult *res, gpointer b
 		}
 
 		g_warning("Mounting of \"%s\" failed: %s", name, error->message);
-		msg = g_strdup_printf(_("Connecting to \"%s\" failed."), name);
-
-		g_signal_emit(backend, signals[OPERATION_FAILED], 0, msg, error->message);
-
+		if (! g_error_matches(error, G_IO_ERROR, G_IO_ERROR_FAILED_HANDLED))
+		{
+			msg = g_strdup_printf(_("Connecting to \"%s\" failed."), name);
+			g_signal_emit(backend, signals[OPERATION_FAILED], 0, msg, error->message);
+			g_free(msg);
+		}
 		g_error_free(error);
 		g_free(name);
-		g_free(msg);
 	}
 	else
 		verbose("Mount finished sucessfully");
@@ -385,7 +460,11 @@ static void unmount_finished_cb(GObject *src, GAsyncResult *res, gpointer backen
 {
 	GError *error = NULL;
 
+#if GLIB_CHECK_VERSION(2, 22, 0)
+	if (! g_mount_unmount_with_operation_finish(G_MOUNT(src), res, &error))
+#else
 	if (! g_mount_unmount_finish(G_MOUNT(src), res, &error))
+#endif
 	{
 		gchar *name, *msg;
 
@@ -410,14 +489,19 @@ static void unmount_finished_cb(GObject *src, GAsyncResult *res, gpointer backen
 }
 
 
-gboolean gigolo_backend_gvfs_mount_volume(GigoloBackendGVFS *backend, gpointer vol)
+gboolean gigolo_backend_gvfs_mount_volume(GigoloBackendGVFS *backend, GtkWindow *window, gpointer vol)
 {
 	g_return_val_if_fail(backend != NULL, FALSE);
+	g_return_val_if_fail(window != NULL, FALSE);
 	g_return_val_if_fail(vol != NULL, FALSE);
 
 	if (! G_IS_MOUNT(vol) && G_IS_VOLUME(vol) && g_volume_can_mount(G_VOLUME(vol)))
 	{
-		g_volume_mount(G_VOLUME(vol), G_MOUNT_MOUNT_NONE, NULL, NULL, volume_mount_finished_cb, backend);
+		GMountOperation *op = gigolo_mount_operation_new(window);
+
+		g_volume_mount(G_VOLUME(vol), G_MOUNT_MOUNT_NONE, op, NULL, volume_mount_finished_cb, backend);
+
+		g_object_unref(op);
 		return TRUE;
 	}
 
@@ -425,12 +509,23 @@ gboolean gigolo_backend_gvfs_mount_volume(GigoloBackendGVFS *backend, gpointer v
 }
 
 
-void gigolo_backend_gvfs_unmount_mount(GigoloBackendGVFS *backend, gpointer mount)
+void gigolo_backend_gvfs_unmount_mount(GigoloBackendGVFS *backend, gpointer mount, GtkWindow *parent)
 {
+#if GLIB_CHECK_VERSION(2, 22, 0)
+	GMountOperation *op;
+#endif
+
 	g_return_if_fail(backend != NULL);
 	g_return_if_fail(mount != NULL);
 
+#if GLIB_CHECK_VERSION(2, 22, 0)
+	op = gigolo_mount_operation_new(parent);
+	g_mount_unmount_with_operation(
+		G_MOUNT(mount), G_MOUNT_UNMOUNT_NONE, op, NULL, unmount_finished_cb, backend);
+	g_object_unref(op);
+#else
 	g_mount_unmount(G_MOUNT(mount), G_MOUNT_UNMOUNT_NONE, NULL, unmount_finished_cb, backend);
+#endif
 }
 
 
@@ -446,7 +541,7 @@ static void mount_ready_cb(GFile *location, GAsyncResult *res, MountInfo *mi)
 	if (error != NULL && ! g_error_matches(error, G_IO_ERROR, G_IO_ERROR_ALREADY_MOUNTED))
 	{
 		gchar *msg = g_strdup_printf(_("Connecting to \"%s\" failed."), uri);
-		if (mi->show_errors)
+		if (mi->show_errors && ! g_error_matches(error, G_IO_ERROR, G_IO_ERROR_FAILED_HANDLED))
 			g_signal_emit(mi->self, signals[OPERATION_FAILED], 0, msg, error->message);
 		else
 			verbose("%s (%s)", msg, error->message);
@@ -485,6 +580,7 @@ void gigolo_backend_gvfs_mount_uri(GigoloBackendGVFS *backend, const gchar *uri,
 		(GAsyncReadyCallback) mount_ready_cb, mi);
 
 	g_object_unref(file);
+	g_object_unref(op);
 }
 
 
@@ -496,105 +592,79 @@ gchar *gigolo_backend_gvfs_get_volume_identifier(gpointer volume)
 }
 
 
-gchar **gigolo_backend_gvfs_get_smb_shares_from_uri(const gchar *uri)
+static gboolean browse_network_ready_cb(gpointer backend)
 {
-	gchar **shares = NULL;
-	GList *l, *shares_list = NULL;
-	GFile *file;
-	GFileInfo *info;
+	GigoloBackendGVFSPrivate *priv;
+	priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(backend);
+
+	g_return_val_if_fail(backend != NULL, FALSE);
+
+	if (priv->browse_counter <= 0)
+	{
+		g_signal_emit(backend, signals[BROWSE_NETWORK_FINISHED], 0);
+		verbose("Browse Network finished");
+		return FALSE;
+	}
+	return TRUE;
+}
+
+
+static void browse_network_mount_ready_cb(GFile *location, GAsyncResult *res, BrowseData *bd)
+{
+	gboolean success;
 	GError *error = NULL;
-	GFileEnumerator *e;
+	GigoloBackendGVFSPrivate *priv;
 
-	g_return_val_if_fail(uri != NULL, NULL);
+	g_return_if_fail(bd != NULL);
+	g_return_if_fail(bd->self != NULL);
 
-	verbose("Querying \"%s\" for available shares", uri);
+	success = g_file_mount_enclosing_volume_finish(location, res, &error);
 
-	file = g_file_new_for_uri(uri);
-
-	e = g_file_enumerate_children(file,
-		G_FILE_ATTRIBUTE_MOUNTABLE_CAN_MOUNT "," G_FILE_ATTRIBUTE_STANDARD_NAME,
-		G_FILE_QUERY_INFO_NONE, NULL, &error);
+	priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(bd->self);
+	priv->browse_counter--;
 
 	if (error != NULL)
 	{
-		verbose("%s: %s", G_STRFUNC, error->message);
+		verbose("%s (%s)", G_STRFUNC, error->message);
 		g_error_free(error);
+
+		/* If we are looking only for shares, we need to emit the finished signal otherwise the
+		 * caller will never know we are done. */
+		if (bd->browse_func == browse_host_real)
+			g_signal_emit(bd->self, signals[BROWSE_HOST_FINISHED], 0, NULL);
 	}
 	else
 	{
-		guint i, len;
-
-		while ((info = g_file_enumerator_next_file(e, NULL, NULL)) != NULL)
-		{
-			shares_list = g_list_append(shares_list, g_strdup(g_file_info_get_name(info)));
-
-			g_object_unref(info);
-		}
-		g_object_unref(e);
-
-		len = g_list_length(shares_list);
-		if (len > 0)
-		{
-			i = 0;
-		shares = g_new(gchar*, len + 1);
-
-		for (l = shares_list; l != NULL; l = g_list_next(l))
-		{
-			shares[i] = l->data;
-			i++;
-		}
-		shares[i] = NULL;
-		}
-		g_list_free(shares_list);
+		bd->browse_func(bd);
 	}
-
-	g_object_unref(file);
-
-	return shares;
 }
 
 
-gchar **gigolo_backend_gvfs_get_smb_shares(const gchar *hostname, const gchar *user, const gchar *domain)
+static void browse_network_real(BrowseData *bd)
 {
-	gchar **shares = NULL;
-	gchar *uri;
-
-	g_return_val_if_fail(hostname != NULL, NULL);
-
-	uri = g_strdup_printf("smb://%s%s%s%s%s/",
-		(NZV(domain)) ? domain : "",
-		(NZV(domain)) ? ";" : "",
-		(NZV(user)) ? user : "",
-		(NZV(user) || NZV(domain)) ? "@" : "",
-		hostname);
-
-	shares = gigolo_backend_gvfs_get_smb_shares_from_uri(uri);
-	g_free(uri);
-
-	return shares;
-}
-
-
-GigoloHostUri **gigolo_backend_gvfs_browse_network(const gchar *uri)
-{
-	GigoloHostUri *h, **hosts = NULL;
-	GList *l, *hosts_list = NULL;
+	GigoloBackendGVFS *backend;
+	GigoloBackendGVFSPrivate *priv;
 	GFile *file;
 	GFileInfo *info;
 	GError *error = NULL;
 	GFileEnumerator *e;
+	GtkTreeStore *store;
+	GtkWindow *parent;
+	gint mode;
 
-	/* If a uri is passed, we assume this is an URI pointing to a workgroup like
-	 * "smb://WORKGROUP/", if passed NULL, we use "smb://" to search for workgroups */
-	if (uri == NULL)
-		uri = "smb://";
+	g_return_if_fail(bd != NULL);
+	g_return_if_fail(bd->self != NULL);
 
-	verbose("Querying \"%s\" for available groups/hosts", uri);
+	store = bd->store;
+	mode = bd->mode;
+	parent = bd->parent;
+	backend = bd->self;
+	priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(backend);
+	priv->browse_counter++;
 
-	file = g_file_new_for_uri(uri);
+	file = g_file_new_for_uri(bd->uri);
 
 	e = g_file_enumerate_children(file,
-		G_FILE_ATTRIBUTE_MOUNTABLE_CAN_MOUNT ","
 		G_FILE_ATTRIBUTE_STANDARD_TARGET_URI ","
 		G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME ","
 		G_FILE_ATTRIBUTE_STANDARD_ICON,
@@ -602,56 +672,197 @@ GigoloHostUri **gigolo_backend_gvfs_browse_network(const gchar *uri)
 
 	if (error != NULL)
 	{
-		verbose("%s: %s", G_STRFUNC, error->message);
-		g_error_free(error);
+		if (g_error_matches(error, G_IO_ERROR, G_IO_ERROR_NOT_MOUNTED))
+		{
+			GMountOperation *op = gigolo_mount_operation_new(bd->parent);
+			/* if the URI wasn't mounted yet, mount it and try again from the mount ready callback */
+			g_file_mount_enclosing_volume(file, G_MOUNT_MOUNT_NONE, op, NULL,
+				(GAsyncReadyCallback) browse_network_mount_ready_cb, bd);
+
+			g_error_free(error);
+			g_object_unref(file);
+			g_object_unref(op);
+			return;
+		}
+		else
+		{
+			verbose("%s: %s", G_STRFUNC, error->message);
+			g_error_free(error);
+		}
 	}
 	else
 	{
-		guint i, len;
-		const gchar *h_uri;
+		const gchar *uri;
+		GtkTreeIter iter, parent_iter_tmp;
+		GtkTreeIter *parent_iter;
+		gint child_mode;
+
+		verbose("Querying \"%s\" for available groups/hosts/shares", bd->uri);
 
 		while ((info = g_file_enumerator_next_file(e, NULL, NULL)) != NULL)
 		{
-			h_uri = g_file_info_get_attribute_string(info, G_FILE_ATTRIBUTE_STANDARD_TARGET_URI);
-			if (h_uri != NULL && g_str_has_prefix(h_uri, "smb://"))
+			uri = g_file_info_get_attribute_string(info, G_FILE_ATTRIBUTE_STANDARD_TARGET_URI);
+			if (uri != NULL)
 			{
-				h = g_new(GigoloHostUri, 1);
-				h->name = g_strdup(g_file_info_get_display_name(info));
-				h->uri = g_strdup(h_uri);
-				h->icon = g_object_ref(g_file_info_get_icon(info));
-				hosts_list = g_list_append(hosts_list, h);
+				/* set parent item */
+				if (bd->parent_path != NULL)
+				{
+					gtk_tree_model_get_iter(GTK_TREE_MODEL(store), &parent_iter_tmp, bd->parent_path);
+					parent_iter = &parent_iter_tmp;
+				}
+				else
+					parent_iter = NULL;
+
+				/* insert the item into the tree store */
+				gtk_tree_store_insert_with_values(store, &iter, parent_iter, -1,
+					GIGOLO_BROWSE_NETWORK_COL_URI, uri,
+					GIGOLO_BROWSE_NETWORK_COL_NAME, g_file_info_get_display_name(info),
+					GIGOLO_BROWSE_NETWORK_COL_ICON, g_file_info_get_icon(info),
+					GIGOLO_BROWSE_NETWORK_COL_CAN_MOUNT, (mode == BROWSE_MODE_SHARES),
+					-1);
+
+				/* set mode for the next level or stop recursion */
+				switch (mode)
+				{
+					case BROWSE_MODE_DOMAINS:
+						child_mode = BROWSE_MODE_HOSTS;
+						break;
+					case BROWSE_MODE_HOSTS:
+						child_mode = BROWSE_MODE_SHARES;
+						break;
+					default:
+						child_mode = BROWSE_MODE_INVALID;
+				}
+
+				/* setup child data and fire it */
+				if (child_mode != BROWSE_MODE_INVALID)
+				{
+					BrowseData *bd_child = g_new0(BrowseData, 1);
+					bd_child->mode = child_mode;
+					bd_child->uri = g_strdup(uri);
+					bd_child->store = store;
+					bd_child->self = backend;
+					bd_child->parent = parent;
+					bd_child->parent_path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);
+					bd_child->browse_func = browse_network_real;
+
+					/* recurse into the next level */
+					browse_network_real(bd_child);
+				}
 			}
 			g_object_unref(info);
 		}
 		g_object_unref(e);
-
-		len = g_list_length(hosts_list);
-		if (len > 0)
-		{
-			i = 0;
-		hosts = g_new(GigoloHostUri*, len + 1);
-
-		for (l = hosts_list; l != NULL; l = g_list_next(l))
-		{
-			hosts[i] = (GigoloHostUri*) l->data;
-			i++;
-		}
-		hosts[i] = NULL;
-		}
-		g_list_free(hosts_list);
 	}
-	g_object_unref(file);
+	priv->browse_counter--;
 
-	return hosts;
+	g_object_unref(file);
+	g_free(bd->uri);
+	gtk_tree_path_free(bd->parent_path);
+	g_free(bd);
 }
 
 
-gpointer gigolo_backend_gvfs_get_share_icon(void)
+void gigolo_backend_gvfs_browse_network(GigoloBackendGVFS *backend, GtkWindow *parent, GtkTreeStore *store)
 {
-	if (gtk_check_version(2, 14, 0) == NULL)
-	return g_themed_icon_new("folder-remote");
+	BrowseData *bd = g_new0(BrowseData, 1);
+	GigoloBackendGVFSPrivate *priv;
+
+	g_return_if_fail(backend != NULL);
+
+	bd->mode = BROWSE_MODE_DOMAINS;
+	bd->parent_path = NULL;
+	bd->parent = parent;
+	bd->store = store;
+	bd->uri = g_strdup("smb://");
+	bd->self = backend;
+	bd->browse_func = browse_network_real;
+
+	priv = GIGOLO_BACKEND_GVFS_GET_PRIVATE(backend);
+	priv->browse_counter = 0;
+
+	browse_network_real(bd);
+
+	/* When we are here, we initiated the network browsing. Then check the 'browse_counter' and
+	 * once it is 0 again, we are probably done with browsing. */
+	g_timeout_add(250, browse_network_ready_cb, backend);
+}
+
+
+static void browse_host_real(BrowseData *bd)
+{
+	GFile *file;
+	GFileInfo *info;
+	GError *error = NULL;
+	GFileEnumerator *e;
+	GSList *list = NULL;
+
+	g_return_if_fail(bd != NULL);
+	g_return_if_fail(bd->self != NULL);
+
+	file = g_file_new_for_uri(bd->uri);
+
+	e = g_file_enumerate_children(file,
+		G_FILE_ATTRIBUTE_STANDARD_NAME,
+		G_FILE_QUERY_INFO_NONE, NULL, &error);
+
+	if (error != NULL)
+	{
+		if (g_error_matches(error, G_IO_ERROR, G_IO_ERROR_NOT_MOUNTED))
+		{
+			GMountOperation *op = gigolo_mount_operation_new(bd->parent);
+			/* if the URI wasn't mounted yet, mount it and try again from the mount ready callback */
+			g_file_mount_enclosing_volume(file, G_MOUNT_MOUNT_NONE, op, NULL,
+				(GAsyncReadyCallback) browse_network_mount_ready_cb, bd);
+
+			g_error_free(error);
+			g_object_unref(file);
+			g_object_unref(op);
+			return;
+		}
+		else
+		{
+			verbose("%s: %s", G_STRFUNC, error->message);
+			g_error_free(error);
+		}
+	}
 	else
-		return NULL;
+	{
+		verbose("Querying \"%s\" for available shares", bd->uri);
+
+		while ((info = g_file_enumerator_next_file(e, NULL, NULL)) != NULL)
+		{
+			list = g_slist_append(list, g_strdup(g_file_info_get_name(info)));
+			g_object_unref(info);
+		}
+		g_object_unref(e);
+	}
+
+	/* propagate our results */
+	g_signal_emit(bd->self, signals[BROWSE_HOST_FINISHED], 0, list);
+
+	g_slist_foreach(list, (GFunc) g_free, NULL);
+	g_slist_free(list);
+
+	g_object_unref(file);
+	g_free(bd->uri);
+	g_free(bd);
+}
+
+
+void gigolo_backend_gvfs_browse_host(GigoloBackendGVFS *backend, GtkWindow *parent, const gchar *hostname)
+{
+	BrowseData *bd = g_new0(BrowseData, 1);
+
+	g_return_if_fail(backend != NULL);
+	g_return_if_fail(NZV(hostname));
+
+	bd->uri = g_strdup_printf("smb://%s", hostname);
+	bd->self = backend;
+	bd->parent = parent;
+	bd->browse_func = browse_host_real;
+
+	browse_host_real(bd);
 }
 
 
