@@ -25,7 +25,6 @@
 
 #include "bookmark.h"
 #include "common.h"
-#include "main.h"
 
 
 typedef struct _GigoloBookmarkPrivate			GigoloBookmarkPrivate;
@@ -38,6 +37,8 @@ struct _GigoloBookmarkPrivate
 	gchar	*name;
 	gchar	*scheme;
 	gchar	*host;
+	gchar	*folder;
+	gchar	*path;
 	gchar	*domain;
 	gchar	*share;
 	guint	 port;
@@ -53,13 +54,15 @@ static void gigolo_bookmark_finalize  		(GObject *object);
 G_DEFINE_TYPE(GigoloBookmark, gigolo_bookmark, G_TYPE_OBJECT);
 
 
-static void bookmark_clear(GigoloBookmark *self)
+void gigolo_bookmark_bookmark_clear(GigoloBookmark *bookmark)
 {
-	GigoloBookmarkPrivate *priv = GIGOLO_BOOKMARK_GET_PRIVATE(self);
+	GigoloBookmarkPrivate *priv = GIGOLO_BOOKMARK_GET_PRIVATE(bookmark);
 
 	g_free(priv->name);
 	g_free(priv->scheme);
 	g_free(priv->host);
+	g_free(priv->folder);
+	g_free(priv->path);
 	g_free(priv->domain);
 	g_free(priv->share);
 	g_free(priv->user);
@@ -69,6 +72,8 @@ static void bookmark_clear(GigoloBookmark *self)
 	priv->host = NULL;
 	priv->port = 0;
 	priv->domain = NULL;
+	priv->path = NULL;
+	priv->folder = NULL;
 	priv->share = NULL;
 	priv->user = NULL;
 
@@ -90,17 +95,17 @@ static void gigolo_bookmark_class_init(GigoloBookmarkClass *klass)
 
 static void gigolo_bookmark_finalize(GObject *object)
 {
-	bookmark_clear(GIGOLO_BOOKMARK(object));
+	gigolo_bookmark_bookmark_clear(GIGOLO_BOOKMARK(object));
 
 	G_OBJECT_CLASS(gigolo_bookmark_parent_class)->finalize(object);
 }
 
 
-static gboolean parse_uri(GigoloBookmark *bm, const gchar *uri)
+gboolean gigolo_bookmark_parse_uri(GigoloBookmark *bookmark, const gchar *uri)
 {
-	gchar *s, *t, *x, *end;
+	gchar *s, *t, *x, *end, *tmp;
 	guint l;
-	GigoloBookmarkPrivate *priv = GIGOLO_BOOKMARK_GET_PRIVATE(bm);
+	GigoloBookmarkPrivate *priv = GIGOLO_BOOKMARK_GET_PRIVATE(bookmark);
 
 	priv->scheme = g_uri_parse_scheme(uri);
 
@@ -108,7 +113,7 @@ static gboolean parse_uri(GigoloBookmark *bm, const gchar *uri)
 	if (priv->scheme == NULL || s == NULL)
 	{
 		verbose("Error parsing URI '%s' while reading URI scheme", uri);
-		bookmark_clear(bm);
+		gigolo_bookmark_bookmark_clear(bookmark);
 		return FALSE;
 	}
 	s += 3;
@@ -136,24 +141,35 @@ static gboolean parse_uri(GigoloBookmark *bm, const gchar *uri)
 		if (l == 0)
 		{
 			verbose("Error parsing URI '%s' while reading username", uri);
-			bookmark_clear(bm);
+			gigolo_bookmark_bookmark_clear(bookmark);
 			return FALSE;
 		}
-		priv->user = g_strndup(s, l);
+		tmp = g_strndup(s, l);
+		/* look for a domain in the username, e.g. "mydomain;myuser" */
+		if (strchr(tmp, ';') != NULL)
+		{
+			gchar **fields = g_strsplit(tmp, ";", 2);
+			priv->domain = fields[0];
+			priv->user = fields[1];
+
+			g_free(fields); /* we free only the array, not the elements */
+		}
+		else
+			priv->user = tmp;
 	}
 
 	/* find hostname */
 	s = (t) ? t + 1 : s;
 	if (*s == '[') /* obex://[00:12:D1:94:1B:28]/ or http://[1080:0:0:0:8:800:200C:417A]/index.html */
 	{
-		gchar *hostend, *tmp;
+		gchar *hostend;
 
 		s++; /* skip the found '[' */
 		hostend = strchr(s, ']');
 		if (! hostend || hostend > end)
 		{
 			verbose("Error parsing URI '%s', missing ']'", uri);
-			bookmark_clear(bm);
+			gigolo_bookmark_bookmark_clear(bookmark);
 			return FALSE;
 		}
 		l = 0;
@@ -184,8 +200,6 @@ static gboolean parse_uri(GigoloBookmark *bm, const gchar *uri)
 	t = strchr(s, ':');
 	if (t != NULL)
 	{
-		gchar *tmp;
-
 		t++; /* skip the found ':' */
 		l = 0;
 		x = t;
@@ -200,8 +214,20 @@ static gboolean parse_uri(GigoloBookmark *bm, const gchar *uri)
 		priv->port = (guint) atoi(tmp);
 		g_free(tmp);
 	}
-	if (NZV(end))
-		priv->share = g_strdup(end + 1);
+	/* TODO handle WebDav path parts */
+	if (NZV(end) && *end == '/' && gigolo_str_equal("smb", priv->scheme))
+	{
+		end++; /* skip the slash */
+
+		l = 0;
+		x = end;
+		while (*x != '\0' && *x != '/')
+		{
+			l++; /* count the len of the share name */
+			x++;
+		}
+		priv->share = g_strndup(end, l);
+	}
 
 	return TRUE;
 }
@@ -209,7 +235,7 @@ static gboolean parse_uri(GigoloBookmark *bm, const gchar *uri)
 
 static void gigolo_bookmark_init(GigoloBookmark *self)
 {
-	bookmark_clear(self);
+	gigolo_bookmark_bookmark_clear(self);
 }
 
 
@@ -225,7 +251,7 @@ GigoloBookmark *gigolo_bookmark_new_from_uri(const gchar *name, const gchar *uri
 	GigoloBookmarkPrivate *priv = GIGOLO_BOOKMARK_GET_PRIVATE(bm);
 
 	gigolo_bookmark_set_name(bm, name);
-	if (! parse_uri(bm, uri))
+	if (! gigolo_bookmark_parse_uri(bm, uri))
 		priv->is_valid = FALSE;
 
 	return bm;
@@ -245,13 +271,15 @@ void gigolo_bookmark_clone(GigoloBookmark *dst, const GigoloBookmark *src)
 	priv_src = GIGOLO_BOOKMARK_GET_PRIVATE(src);
 
 	/* free existing strings and data */
-	bookmark_clear(dst);
+	gigolo_bookmark_bookmark_clear(dst);
 
 	/* copy from src to dst */
 	priv_dst->name = g_strdup(priv_src->name);
 	priv_dst->host = g_strdup(priv_src->host);
 	priv_dst->scheme = g_strdup(priv_src->scheme);
 	priv_dst->domain = g_strdup(priv_src->domain);
+	priv_dst->folder = g_strdup(priv_src->folder);
+	priv_dst->path = g_strdup(priv_src->path);
 	priv_dst->share = g_strdup(priv_src->share);
 	priv_dst->user = g_strdup(priv_src->user);
 	priv_dst->port = priv_src->port;
@@ -262,6 +290,7 @@ static gchar *gigolo_bookmark_get_uri_real(GigoloBookmark *bookmark, gboolean es
 {
 	GigoloBookmarkPrivate *priv = GIGOLO_BOOKMARK_GET_PRIVATE(bookmark);
 	gchar *result;
+	const gchar *domain = NULL;
 	gchar *user = NULL;
 	gchar *port = NULL;
 
@@ -278,12 +307,18 @@ static gchar *gigolo_bookmark_get_uri_real(GigoloBookmark *bookmark, gboolean es
 	else
 		user = g_strdup(priv->user);
 
-	result = g_strdup_printf("%s://%s%s%s%s/%s%s",
+	domain = gigolo_bookmark_get_domain(bookmark);
+
+	result = g_strdup_printf("%s://%s%s%s%s%s%s/%s%s%s%s",
 		priv->scheme,
+		(NZV(domain)) ? domain : "",
+		(NZV(domain)) ? ";" : "",
 		(NZV(user)) ? user : "",
-		(NZV(user)) ? "@" : "",
+		(NZV(user) || NZV(domain)) ? "@" : "",
 		priv->host,
 		(port) ? port : "",
+		(NZV(priv->path)) ? priv->path : "",
+		(NZV(priv->path)) ? "/" : "",
 		(NZV(priv->share)) ? priv->share : "",
 		(NZV(priv->share)) ? "/" : "");
 
@@ -387,6 +422,50 @@ void gigolo_bookmark_set_host(GigoloBookmark *bookmark, const gchar *host)
 }
 
 
+const gchar *gigolo_bookmark_get_folder(GigoloBookmark *bookmark)
+{
+	g_return_val_if_fail(bookmark != NULL, NULL);
+
+	return GIGOLO_BOOKMARK_GET_PRIVATE(bookmark)->folder;
+}
+
+
+void gigolo_bookmark_set_folder(GigoloBookmark *bookmark, const gchar *folder)
+{
+	GigoloBookmarkPrivate *priv;
+
+	g_return_if_fail(bookmark != NULL);
+	g_return_if_fail(folder != NULL);
+
+	priv = GIGOLO_BOOKMARK_GET_PRIVATE(bookmark);
+
+	g_free(priv->folder);
+	priv->folder = g_strdup(folder);
+}
+
+
+const gchar *gigolo_bookmark_get_path(GigoloBookmark *bookmark)
+{
+	g_return_val_if_fail(bookmark != NULL, NULL);
+
+	return GIGOLO_BOOKMARK_GET_PRIVATE(bookmark)->path;
+}
+
+
+void gigolo_bookmark_set_path(GigoloBookmark *bookmark, const gchar *path)
+{
+	GigoloBookmarkPrivate *priv;
+
+	g_return_if_fail(bookmark != NULL);
+	g_return_if_fail(path != NULL);
+
+	priv = GIGOLO_BOOKMARK_GET_PRIVATE(bookmark);
+
+	g_free(priv->path);
+	priv->path = g_strdup(path);
+}
+
+
 guint gigolo_bookmark_get_port(GigoloBookmark *bookmark)
 {
 	g_return_val_if_fail(bookmark != NULL, 0);
