@@ -24,9 +24,13 @@
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
 
-#include "common.h"
-#include "main.h"
+#ifdef GDK_WINDOWING_X11
+ #include <gdk/gdkx.h>
+#endif
 
+#include "common.h"
+
+gboolean verbose_mode;
 
 const gchar *gigolo_get_application_icon_name(void)
 {
@@ -80,14 +84,17 @@ const gchar *gigolo_describe_scheme(const gchar *scheme)
 	else if (gigolo_str_equal(scheme, "sftp"))
 		return _("SSH");
 	else if (gigolo_str_equal(scheme, "obex"))
-		/* TODO find something better */
-		return _("OBEX");
+		return _("Obex");
 	else if (gigolo_str_equal(scheme, "dav"))
 		return _("WebDAV");
 	else if (gigolo_str_equal(scheme, "davs"))
 		return _("WebDAV (secure)");
 	else if (gigolo_str_equal(scheme, "network"))
 		return _("Network");
+	else if (gigolo_str_equal(scheme, "archive"))
+		return _("Archive");
+	else if (gigolo_str_equal(scheme, "gphoto2"))
+		return _("Photos");
 	else if (gigolo_str_equal(scheme, "custom"))
 		return _("Custom Location");
 
@@ -118,20 +125,27 @@ gboolean gigolo_is_desktop_xfce(void)
 
 	if (check)
 	{
-		gint result;
-		gchar *out = NULL;
-		gboolean success;
-
-		success = g_spawn_command_line_sync("xprop -root _DT_SAVE_MODE", &out, NULL, &result, NULL);
-		if (success && result == 0 && out != NULL && strstr(out, "xfce4") != NULL)
+		/* Are we running in Xfce? */
+		GdkDisplay *display = gdk_display_get_default();
+		Display *xdisplay = GDK_DISPLAY_XDISPLAY(display);
+		Window root_window = RootWindow(xdisplay, 0);
+		Atom save_mode_atom = gdk_x11_get_xatom_by_name("_DT_SAVE_MODE");
+		Atom actual_type;
+		gint actual_format;
+		gulong n_items, bytes;
+		gchar *value;
+		gint status = XGetWindowProperty(xdisplay, root_window, save_mode_atom, 0, (~0L),
+			False, AnyPropertyType, &actual_type, &actual_format, &n_items, &bytes,
+			(guchar**) &value);
+		if (status == Success)
 		{
-			is_xfce = TRUE;
+			if (n_items == 6 && !strncmp (value, "xfce4", 6))
+				is_xfce = TRUE;
+			XFree(value);
 		}
-		g_free(out);
-
 		check = FALSE;
 	}
-    return is_xfce;
+	return is_xfce;
 }
 
 
@@ -174,3 +188,29 @@ void gigolo_show_uri(const gchar *uri)
 	g_free(cmd);
 	g_free(open_cmd);
 }
+
+
+#ifdef DEBUG
+void debug(gchar const *format, ...)
+{
+	va_list args;
+	va_start(args, format);
+	g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, format, args);
+	va_end(args);
+}
+#endif
+
+
+void verbose(gchar const *format, ...)
+{
+#ifndef DEBUG
+	if (verbose_mode)
+#endif
+	{
+		va_list args;
+		va_start(args, format);
+		g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, format, args);
+		va_end(args);
+	}
+}
+
